/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla XHTML + Voice support.
 *
 * The Initial Developer of the Original Code is IBM Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Chris Cross <xcross@us.ibm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <stdio.h>
#include "nsString.h"
#include "nsCOMPtr.h"
#include "nsIConsoleService.h"
#include "nsContentUtils.h"
#include "nsIJSContextStack.h"
#include "nsIServiceManager.h"
#include "nsIScriptGlobalObject.h"
#include "nsIScriptObjectPrincipal.h"
#include "nsIScriptSecurityManager.h"
#include "nsIPrincipal.h"
#include "jsapi.h"
#include "jsinterp.h"
#include "jscntxt.h"

#include "nsMozillaScope.h"
#include <wchar.h>
#pragma warning(disable:4800)

#define DEBUGTRACE
#include "debugtrace.h"


int MozillaScope::m_errorRc = 0;

static JSClass objClass = {
    "Object", 0,
    JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   JS_FinalizeStub
};

class AutoPushJSContext
{
public:
  AutoPushJSContext(JSContext *cx);

  ~AutoPushJSContext();

  nsresult ResultOfPush() { return mPushResult; };

private:
  nsCOMPtr<nsIJSContextStack> mContextStack;
  JSContext*                  mContext;
  JSStackFrame                mFrame;
  nsresult                    mPushResult;
};

AutoPushJSContext::AutoPushJSContext(JSContext *cx) 
  : mContext(cx), mPushResult(NS_OK)
{
  nsCOMPtr<nsIJSContextStack> contextStack =
    do_GetService("@mozilla.org/js/xpc/ContextStack;1");

  JSContext* currentCX;
  if(contextStack &&
     // Don't push if the current context is already on the stack.
     (NS_FAILED(contextStack->Peek(&currentCX)) ||
      cx != currentCX) )
  {
    if (NS_SUCCEEDED(contextStack->Push(cx)))
    {
      // Leave the reference in mContextStack to
      // indicate that we need to pop it in our dtor.
      mContextStack.swap(contextStack);
    }
  }

  nsCOMPtr<nsIScriptSecurityManager> secMan(do_GetService(NS_SCRIPTSECURITYMANAGER_CONTRACTID, &mPushResult));

  if (NS_FAILED(mPushResult))
    return;

  nsCOMPtr<nsIPrincipal> principal;
  mPushResult = secMan->GetPrincipalFromContext(cx, getter_AddRefs(principal));

  if (NS_FAILED(mPushResult))
  {
    JS_ReportError(cx, "failed to get a principal");
    return;
  }

  // See if JavaScript is enabled for the current window
  PRBool jsEnabled = PR_FALSE;
  mPushResult = secMan->CanExecuteScripts(cx, principal, &jsEnabled);
  if (!jsEnabled)
    mPushResult = NS_ERROR_FAILURE;

  memset(&mFrame, 0, sizeof(mFrame));

  if (NS_SUCCEEDED(mPushResult))
  {
    // See if there are any scripts on the stack.
    // If not, we need to add a dummy frame with a principal.
    
    PRBool hasScript = PR_FALSE;
    JSStackFrame* tempFP = cx->fp;
    while (tempFP)
    {
      if (tempFP->script)
      {
        hasScript = PR_TRUE;
        break;
      }
      tempFP = tempFP->down;
    };

    if (!hasScript)
    {
      JSPrincipals* jsprinc;
      principal->GetJSPrincipals(cx, &jsprinc);

      mFrame.script = JS_CompileScriptForPrincipals(cx, JS_GetGlobalObject(cx), jsprinc, "", 0, "", 1);
      JSPRINCIPALS_DROP(cx, jsprinc);

      if (mFrame.script)
      {
        mFrame.down = cx->fp;
        cx->fp = &mFrame;
      }
      else
        mPushResult = NS_ERROR_OUT_OF_MEMORY;
    }
  }
}

AutoPushJSContext::~AutoPushJSContext()
{
  if (mContextStack)
    mContextStack->Pop(nsnull);

  if (mFrame.script)
    mContext->fp = mFrame.down;
}

VxmScope * 
MozillaScope::createScope(VxmContext context, GlobalScope globalScope)
{
  FUNCTION_CLASS_TRACE( MozillaScope::createScope);

  VxmScope *scope = (VxmScope *)(new MozillaScope((JSContext*)context, (nsIDocument*)globalScope));
  return scope;
}

/**Ctor
 * @param context pointer to Spider Monkey thread context
 * @param gscope  pointer to Global Scope
 */
MozillaScope::MozillaScope(JSContext* context, nsIDocument* aDoc) 
  : VxmScope(), m_Doc(aDoc), m_Context(context), m_Object(NULL)
{
  FUNCTION_CLASS_TRACE( MozillaScope::MozillaScope );
  printf("New Mozilla Scope with nsIDocument * %0.8x\n", m_Doc);
  // QI the doc here?
  NS_ASSERTION(aDoc, "Null document");
  NS_ASSERTION(context, "Null context");

  // Get the global object and context from the Document 
  nsIScriptGlobalObject *sgo = m_Doc->GetScriptGlobalObject();
  m_Object = sgo->GetGlobalJSObject();

  // Set the SpiderMonkey error reporter
  JS_SetErrorReporter(m_Context, errorReporter);

  // Push global scope onto the scope chain
  SMObject* objectNode = new SMObject(m_Object);
  m_scopeChain.InsertElementAt(objectNode, 0);
}

/**Dtor
 */
MozillaScope::~MozillaScope()
{
  FUNCTION_CLASS_TRACE( MozillaScope::~MozillaScope );

  if (m_Object)
  {

    // Clear scope chain of all objects
    for (PRInt32 i = 0; i < m_scopeChain.Count(); i++)
    {
      JSObject* obj = ((SMObject*)(m_scopeChain[i]))->m_object;
      if (obj != m_Object) // Don't remove global scope properties
      {
        JS_ClearScope(m_Context, obj);
        JSVAL_TO_GCTHING(OBJECT_TO_JSVAL(obj));
        delete (SMObject*) m_scopeChain[i];
      }
    }

    m_scopeChain.Clear();
  }
}

/**ErrorReporter error handler
 * @param context SpiderMonkey context
 * @param msg     Error message
 * @param report  JSErrorReport structure
 */
void MozillaScope::errorReporter(JSContext* context, 
                    const char* msg, JSErrorReport* report)
{
  if (report != NULL) {
    char * buf = new char[2*strlen(msg)];
    nsCOMPtr<nsIConsoleService> console = do_GetService( NS_CONSOLESERVICE_CONTRACTID );
    NS_ASSERTION(console, "Failed to get console service.");

    m_errorRc = report->errorNumber;
    printf("error rc=%d msg=\nSTART\n%s\nEND\n", m_errorRc, msg);
    sprintf(buf, "JavaScript error in VoiceXML form rc=%d msg=\n%s\n", m_errorRc, msg);
    nsCAutoString s(buf);
    console->LogStringMessage(NS_ConvertUTF8toUTF16(s).get());
    printf("filename=%s\nlineno=%d\n", report->filename, report->lineno);
    sprintf(buf, "filename=%s\nlineno=%d\n", report->filename, report->lineno);
    s = buf;
    console->LogStringMessage(NS_ConvertUTF8toUTF16(s).get());
  }

}

VxmRc 
MozillaScope::init(nsIDocument *aDoc)
{
  NS_ASSERTION(aDoc, "Null document");

  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  InitScopeEvent * evt = new InitScopeEvent(aDoc);
  if (!evt) 
  {
      return VXM_RC_MEM_ALLOC_ERROR;
   }
 
  // Make the owner NULL since this is a static method
   PL_InitEvent(evt, NULL, MozillaScope::HandleInitEvent, MozillaScope::DestroyInitEvent);
 
  void * result;
   nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) 
   {
     PL_DestroyEvent(evt);
    return VXM_RC_INTERNAL_ERROR;
   }

  //  rc is updated in the event
  return evt->mRc;
}

void * 
MozillaScope::HandleInitEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleInitEvent );

  InitScopeEvent * evt = NS_STATIC_CAST(InitScopeEvent *, aEvent);
  nsIDocument *doc = evt->mDoc;

  uintN lineno = 0;

  // Create VXMLAudioRecording type
  //NS_NAMED_LITERAL_STRING(script,"function VXMLAudioRecording(hndl){ this.handle=hndl;}");
   NS_NAMED_LITERAL_STRING(script,"VXMLAudioRecording = new Function('hndl', 'this.handle=hndl;');");

  // Get the principles and global object
  JSPrincipals *jsPrins = NULL;
  nsIScriptGlobalObject *sgo = doc->GetScriptGlobalObject();
  NS_ASSERTION(sgo, "GetScriptGlobalObject failed, crashing soon."); 

  nsCOMPtr<nsIScriptObjectPrincipal> sop(do_QueryInterface(sgo));
  nsIScriptContext *scrCx = sgo->GetContext();
  JSObject * glob = sgo->GetGlobalJSObject();
  JSContext* cx = (JSContext *) scrCx->GetNativeContext();
  nsIPrincipal *prin = sop->GetPrincipal();
  prin->GetJSPrincipals(cx, &jsPrins);
  NS_ASSERTION(jsPrins, "Unable to get js principals!");

  // Evaluate the script
  // Push the context on the stack in FireFox
  JSBool jsok;
  jsval jsValue;
  nsresult rv;
  nsCOMPtr<nsIJSContextStack> cxstack = 
    do_GetService("@mozilla.org/js/xpc/ContextStack;1");
  rv = cxstack->Push(cx);
  if (NS_FAILED(rv)) {
    NS_ERROR("coudn't push the context on the stack");
    evt->mRc = VXM_RC_INTERNAL_ERROR;
    return (void *)evt;
  }

  JS_BeginRequest(cx);

  // Evaluate the script. 
  jsok = JS_EvaluateUCScriptForPrincipals(cx, glob, jsPrins, script.get(), script.Length(), "x+v script", lineno, &jsValue);

  JS_EndRequest(cx);

  JS_MaybeGC(cx);

  JSContext *acx;
  cxstack->Pop(&acx);
  NS_ASSERTION(acx == acx, "JS contexts didn't match");

  if (!jsok)
  {
    evt->mRc = VXM_RC_INTERNAL_ERROR;
    wprintf(L"Internal javascript error: eval %s \n",script.get());
  } 

  evt->mRc = VXM_RC_OK;
  return (void *)evt;
}

/**Evaluate the EcmaScript contained in script as program text.
 *  Return the result of the evaluation in the result.
 * @param script (in) Program text to compile and execute.
 * @param result (out) Result of script evaluation
 * @param bReleaseObject (in) if true release object if object returned
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED
 */

VxmRc 
MozillaScope::eval(const wchar_t *script, EcmaScriptValue* aValue, bool bReleaseObject)
{

  NS_ASSERTION(script, "Null script");
  NS_ASSERTION(aValue, "Null EcmaScriptValue address");

  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSEvalEvent * evt = new JSEvalEvent(this, (wchar_t *)script, aValue, bReleaseObject);
  if (!evt) 
  {
      return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleEvalEvent, MozillaScope::DestroyEvalEvent);
 
  void * result;
   nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) 
   {
     PL_DestroyEvent(evt);
    return VXM_RC_INTERNAL_ERROR;
   }

  //  aValue is updated directly in the event
  return evt->mRc;
}

void * 
MozillaScope::HandleEvalEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleEvalEvent );

  JSEvalEvent * evt = NS_STATIC_CAST(JSEvalEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);
  uintN lineno = 0;

  JSObject* obj = ((SMObject*)(scope->m_scopeChain[0]))->m_object;


  // gm defect 127366
  // There is a bug with SpiderMonkey setting a variable to undefined 
  // when re-running a voicexml form.  Fix is to strip out 'undefined
  // and replace with calling setSlot with SCOPE_VALUE_UNDEFINED.  In
  // setSlot set the input jsval to JSVAL_VOID.

  //cc Using the Netscape Auto String (painful) 
  nsAutoString s(evt->mScript);
  nsAString::const_iterator eq_start, un_start, end;
  s.BeginReading(eq_start);
  s.BeginReading(un_start);
  s.EndReading(end);
  NS_NAMED_LITERAL_STRING(s_eq, "=");
  NS_NAMED_LITERAL_STRING(s_un, "undefined");
  bool found_eq = FindInReadable(s_eq, eq_start, end);
  s.EndReading(end);
  bool found_un = FindInReadable(s_un, un_start, end);

  if (found_eq && found_un && un_start.get() > eq_start.get())
  {
    EcmaScriptValue esvalue;
    esvalue.type = SCOPE_VALUE_UNDEFINED;
    esvalue.value.object_val = (void*)NULL;
    s.Cut(0, s.BeginReading() - eq_start.get()); //cc check for one-off error
    VxmRc rc = scope->setSlot(s.get(), esvalue);
    return NULL; //cc fix up just getting to compile here
  }

  // Get the priciples
  JSPrincipals *jsPrins = NULL;
  nsIScriptGlobalObject *sgo = scope->m_Doc->GetScriptGlobalObject();
  nsCOMPtr<nsIScriptObjectPrincipal> sop(do_QueryInterface(sgo));
  nsIPrincipal *prin = sop->GetPrincipal();
  prin->GetJSPrincipals(scope->m_Context, &jsPrins);
  NS_ASSERTION(jsPrins, "Unable to get js principals!");

  // Evaluate the script
  // Push the context on the stack in FireFox
  JSBool jsok;
  jsval jsValue;
  nsresult rv;
  nsCOMPtr<nsIJSContextStack> cxstack = 
    do_GetService("@mozilla.org/js/xpc/ContextStack;1");
  rv = cxstack->Push(scope->m_Context);
  if (NS_FAILED(rv)) {
    NS_ERROR("coudn't push the context on the stack");
    return NULL; //cc fix up just getting to compile here VXM_RC_INTERNAL_ERROR;
  }

  JS_BeginRequest(scope->m_Context);

  // Evaluate the script
  jsok = JS_EvaluateUCScriptForPrincipals(scope->m_Context, obj, jsPrins, evt->mScript.get(), evt->mScript.Length(), "x+v script", lineno, &jsValue);

  JS_EndRequest(scope->m_Context);

  JS_MaybeGC(scope->m_Context);

  JSContext *cx;
  cxstack->Pop(&cx);
  NS_ASSERTION(cx == scope->m_Context, "AutoConfig JS contexts didn't match");

  if (!jsok)
  {
    evt->mResult->type = SCOPE_VALUE_UNDEFINED;
    evt->mResult->value.object_val = (void*)NULL;
    evt->mRc = VXM_RC_INTERNAL_ERROR;
    wprintf(L"Internal javascript error: eval %s \n\t rc = %d\n",evt->mScript.get(),scope->m_errorRc);
    return (void *)evt;
  } 

  // Check the type and set the result type and value
  if (JSVAL_IS_VOID(jsValue))
  {
    evt->mResult->type = SCOPE_VALUE_UNDEFINED;
    evt->mResult->value.object_val = (void*)NULL;
  }   
  else if (JSVAL_IS_OBJECT(jsValue))
  {
    JSObject* jobj;
    JS_ValueToObject(scope->m_Context, jsValue, &jobj);
    
    evt->mResult->type = SCOPE_VALUE_OBJECT;
    evt->mResult->value.object_val = jobj;

  }
  else if (JSVAL_IS_BOOLEAN(jsValue))
  {
    evt->mResult->type = SCOPE_VALUE_BOOLEAN;
    evt->mResult->value.bool_val = (JSVAL_TO_BOOLEAN(jsValue) != 0);
  }
  else if (JSVAL_IS_NUMBER(jsValue))
  {
    evt->mResult->type = SCOPE_VALUE_NUMBER;
    JS_ValueToNumber(scope->m_Context, jsValue, &evt->mResult->value.num_val);
  }
  else if (JSVAL_IS_STRING(jsValue))
  {
    evt->mResult->type = SCOPE_VALUE_STRING;
    JSString *jstr = JS_ValueToString(scope->m_Context, jsValue);
    evt->mResult->value.str_val = (wchar_t*)JS_GetStringChars(jstr);
  }
  else
  {
    evt->mResult->type = SCOPE_VALUE_UNDEFINED;
    evt->mResult->value.object_val = (void*)NULL;
  }

  return (void *)evt;
}


/**Enter a new inner scope
 * @param name (in) Name of the scope object
 * @param object (out) scope object created by enter.  Input to object
 *                     is NULL if caller not interested in it.
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED
 */
VxmRc MozillaScope::enter(const wchar_t *name, ESObject* object)
{
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSEnterEvent * evt = new JSEnterEvent(this, (wchar_t *)name, object);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleEnterEvent, MozillaScope::DestroyEnterEvent);
 
   void * result;
   nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
 
   if (NS_FAILED(rv)) {
     PL_DestroyEvent(evt);
    return VXM_RC_INTERNAL_ERROR;
   }

  //The requested object was updated directly in the event
  return evt->mRc;
}

void *
MozillaScope::HandleEnterEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleEnterEvent );


  JSEnterEvent * evt = NS_STATIC_CAST(JSEnterEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  jsval slotVal;

  JSObject* parent = ((SMObject*)(scope->m_scopeChain[0]))->m_object;

  JSObject* jobj = JS_NewObject(scope->m_Context, &objClass, NULL, parent);

  if (jobj != NULL)
  {

    slotVal = OBJECT_TO_JSVAL(jobj);

    nsAutoString slotname;
    slotname.Assign(evt->mName.get(), evt->mName.Length());
    if (slotname.Length() == 0)
    {
      slotname = L"Anonymous";
    }

    // Set the slot value as a spidermonkey property
    if (!JS_SetUCProperty(scope->m_Context,
                parent,
                (jschar*)slotname.get(),
                slotname.Length(),
                &slotVal))
    {
      evt->mRc = VXM_RC_INTERNAL_ERROR;
      return (void *) evt;
    }

    // Push scope object onto the scope chain
    SMObject* objectNode = new SMObject(jobj);
    scope->m_scopeChain.InsertElementAt(objectNode, 0);
  }
  else
  {
    evt->mRc = VXM_SCOPE_RC_NO_OBJECT;
  }

  //Enter may be called with a NULL object for Anonymous scope
  if (evt->mObject)
  {
    *evt->mObject = (ESObject)jobj;
  }

  return (void *)evt;
}

/**Leave the current scope
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_SCOPE_RC_NO_OBJECT
 */
VxmRc MozillaScope::leave()
{
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSLeaveEvent * evt = new JSLeaveEvent(this);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleLeaveEvent, MozillaScope::DestroyLeaveEvent);
 
  void * result;
  nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) 
   {
     PL_DestroyEvent(evt);
    return VXM_RC_INTERNAL_ERROR;
   }
  return evt->mRc;
}

void *
MozillaScope::HandleLeaveEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleLeaveEvent );

  JSCreateObjectEvent * evt = NS_STATIC_CAST(JSCreateObjectEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  VxmRc rc = VXM_RC_OK;
  SMObject * smObj = (SMObject*)scope->m_scopeChain[0]; 
  JSObject* obj = smObj->m_object;

  if (obj && obj != scope->m_Object)
  {
    // Remove all properties associated with object
    JS_ClearScope(scope->m_Context, obj);

    // Remove protection from garbage collection
    JSVAL_TO_GCTHING(OBJECT_TO_JSVAL(obj));

    // Remove object from scope chain
    scope->m_scopeChain.RemoveElementAt(0);
    delete smObj;
    smObj = 0;
  }
  else
  {
    evt->mRc = VXM_SCOPE_RC_NO_OBJECT;
  }

  return (void *)evt;
}

/**Create an unnamed object
 * @param object (out) handle to the created object.  This
 *                  object must be later released by
 *                  calling releaseObject
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED,
 *         VXM_SCOPE_RC_NO_OBJECT
 */
VxmRc MozillaScope::createObject(ESObject* aObject)
{
  NS_ASSERTION(aObject, "Null object");

  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSCreateObjectEvent * evt = new JSCreateObjectEvent(this, aObject);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleCreateObjectEvent, MozillaScope::DestroyCreateObjectEvent);
 
  void * result;
   nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) 
   {
     PL_DestroyEvent(evt);
    return VXM_RC_INTERNAL_ERROR;
  }

  //The requested object was updated directly in the event
  return evt->mRc;
}

void *
MozillaScope::HandleCreateObjectEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleCreateObjectEvent );

  JSCreateObjectEvent * evt = NS_STATIC_CAST(JSCreateObjectEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  VxmRc rc = VXM_RC_OK;

  // not attached to an object, can be garbage collected
  JSObject* jobj = JS_NewObject(scope->m_Context, &objClass, NULL, NULL);
  if (jobj == NULL)
  {
    evt->mRc = VXM_RC_INTERNAL_ERROR;
    return (void *)evt;
  }

  *evt->mObject = (ESObject)jobj;
  return  (void *)evt;
}

/**Create an unnamed array
 * @param array (out) handle to the created array.  This
 *                    array must be later released by
 *                    calling releaseObject.
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED,
 *         VXM_SCOPE_RC_NO_OBJECT
 */
VxmRc MozillaScope::createArrayObject(ESObject* array)
{
  NS_ASSERTION(array, "Null array");

  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSCreateArrayObjectEvent * evt = new JSCreateArrayObjectEvent(this, array);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleCreateArrayObjectEvent, MozillaScope::DestroyCreateArrayObjectEvent);
 
  void * result;
   nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) 
   {
     PL_DestroyEvent(evt);
     return VXM_RC_INTERNAL_ERROR;
   }

  //The requested  array was updated directly in the event
  return evt->mRc;
}

void *
MozillaScope::HandleCreateArrayObjectEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleCreateArrayObjectEvent );

  JSCreateArrayObjectEvent * evt = NS_STATIC_CAST(JSCreateArrayObjectEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  VxmRc rc = VXM_RC_OK;

  // not attached to an object, can be garbage collected
  JSObject* jobj = JS_NewArrayObject(scope->m_Context, 0, NULL);
  if (jobj == NULL)
  {
    evt->mRc = VXM_SCOPE_RC_NO_OBJECT;
    return (void *)evt;
  }

  *evt->mArray = (ESObject)jobj;
  return (void *)evt;
}

/**Release an object or array
 * @param obj (in) handle to the object
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_SCOPE_RC_NO_OBJECT
 */
VxmRc 
MozillaScope::releaseObject(ESObject object)
{
  //Release object is a noop in Mozilla, let GC use ref count 
  return VXM_RC_OK;
}

/**Get the value of a field within an object
 * @param slotName (in) Name of the field
 * @param result (out) returned value of the field
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED,
 *         VXM_SCOPE_RC_NO_OBJECT
 */
VxmRc MozillaScope::getSlot(const wchar_t * slotName, EcmaScriptValue* aValue)
{
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) 
   {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSGetSlotEvent * evt = new JSGetSlotEvent(this, slotName, aValue);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleGetSlotEvent, MozillaScope::DestroyGetSlotEvent);
 
  void * result;
   nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) 
   {
     PL_DestroyEvent(evt);
     return VXM_RC_INTERNAL_ERROR;
  }

  //aValue being updated directly in the event
  return evt->mRc;
}

void *
MozillaScope::HandleGetSlotEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleGetSlotEvent );
    //  LOG_CLASS_STR_FUNCTION( NS_LITERAL_CSTRING("Arguments: slotName = ") + slotName + "\n\t result = " + result)

  JSGetSlotEvent * evt = NS_STATIC_CAST(JSGetSlotEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  // get scope object at the top of the stack
  JSObject* jobj = ((SMObject*)scope->m_scopeChain[0])->m_object;

  // error if object not found
  if (jobj == NULL)
  {
    evt->mRc =  VXM_SCOPE_RC_NO_OBJECT;
    return (void *)evt;
  }

  jsval value;  // Spidermonkey property value

  // Get the Spidermonkey object property
  if (!JS_GetUCProperty(scope->m_Context,
              jobj, 
              (jschar*)evt->mSlotName.get(),             
              evt->mSlotName.Length(),
              &value))
  {
    evt->mValue->type = SCOPE_VALUE_UNDEFINED;
    evt->mValue->value.object_val = (void*)NULL;
    evt->mRc = VXM_RC_INTERNAL_ERROR;
    return (void *)evt;
  }

  // Set the output type and value from the returned value
  if (JSVAL_IS_VOID(value))
  {
    evt->mValue->type = SCOPE_VALUE_UNDEFINED;
    evt->mValue->value.object_val = (void*)NULL;
  }
  else if (JSVAL_IS_OBJECT(value))
  {
    evt->mValue->type = SCOPE_VALUE_OBJECT;
    JSObject *temp = 0;
    JS_ValueToObject(scope->m_Context, value, &temp);
    evt->mValue->value.object_val = temp;
  }
  else if (JSVAL_IS_BOOLEAN(value))
  {
    evt->mValue->type = SCOPE_VALUE_BOOLEAN;
    evt->mValue->value.bool_val = (JSVAL_TO_BOOLEAN(value) != 0);
  }
  else if (JSVAL_IS_NUMBER(value))
  {
    evt->mValue->type = SCOPE_VALUE_NUMBER;
    JS_ValueToNumber(scope->m_Context, value, &evt->mValue->value.num_val);
  }
  else if (JSVAL_IS_STRING(value))
  {
    evt->mValue->type = SCOPE_VALUE_STRING;
    JSString *str = JS_ValueToString(scope->m_Context, value);
    evt->mValue->value.str_val = (wchar_t*)JS_GetStringChars(str);
  }
  else
  {
    evt->mValue->type = SCOPE_VALUE_UNDEFINED;
    evt->mValue->value.object_val = (void*)NULL;
  }

  return (void *)aEvent;
}

/**Get the value of a field within an object
 * @param slotName (in) Name of the field
 * @param result (out) returned value of the field
 * @param object (in) handle to the slot's object.  When
 *                    object is NULL object at top of the
 *                    scope chain used to get the slot.
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED
 */
VxmRc
MozillaScope::getSlot(const wchar_t* slotName, EcmaScriptValue* aResult, ESObject object)
{
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) 
   {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSGetSlotObjectEvent * evt = new JSGetSlotObjectEvent(this, slotName, aResult, object);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleGetSlotObjectEvent, MozillaScope::DestroyGetSlotObjectEvent);
 
  void * result;
   nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) 
   {
     PL_DestroyEvent(evt);
     return VXM_RC_INTERNAL_ERROR;
   }

  // aResult updated in the event with slot value
  return evt->mRc;
}

void *
MozillaScope::HandleGetSlotObjectEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleGetSlotObjectEvent );

  JSGetSlotObjectEvent * evt = NS_STATIC_CAST(JSGetSlotObjectEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  // Use the given scope chain
  JSObject* jobj = (JSObject*)evt->mObject; 
  jsval value;  // Spidermonkey property value

  // Get the Spidermonkey object property
  if (!JS_GetUCProperty(scope->m_Context,
              jobj, 
              (jschar*)evt->mSlotName.get(),             
              evt->mSlotName.Length(), 
              &value))
  {
    evt->mValue->type = SCOPE_VALUE_UNDEFINED;
    evt->mValue->value.object_val = (void*)NULL;
    evt->mRc = VXM_RC_INTERNAL_ERROR;
    return (void *)evt;
  }

  // Set the output type and value from the returned value
  if (JSVAL_IS_VOID(value))
  {
    evt->mValue->type = SCOPE_VALUE_UNDEFINED;
    evt->mValue->value.object_val = (void*)NULL;
  }
  else if (JSVAL_IS_OBJECT(value))
  {
    evt->mValue->type = SCOPE_VALUE_OBJECT;
    JSObject *temp = 0;
    JS_ValueToObject(scope->m_Context, value, &temp);
    evt->mValue->value.object_val = temp;
  }
  else if (JSVAL_IS_BOOLEAN(value))
  {
    evt->mValue->type = SCOPE_VALUE_BOOLEAN;
    evt->mValue->value.bool_val = (JSVAL_TO_BOOLEAN(value) != 0);
  }
  else if (JSVAL_IS_NUMBER(value))
  {
    evt->mValue->type = SCOPE_VALUE_NUMBER;
    JS_ValueToNumber(scope->m_Context, value, &evt->mValue->value.num_val);
  }
  else if (JSVAL_IS_STRING(value))
  {
    evt->mValue->type = SCOPE_VALUE_STRING;
    JSString *str = JS_ValueToString(scope->m_Context, value);
    evt->mValue->value.str_val = (wchar_t*)JS_GetStringChars(str);
  }
  else
  {
    evt->mValue->type = SCOPE_VALUE_UNDEFINED;
    evt->mValue->value.object_val = (void*)NULL;
  }

  return (void *)evt;
}
/**Get the value of an element within an array
 * @param index (in) index number of the element in array
 * @param result (out) returned value of the field
 * @param array (in) handle to the element's array
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED,
 *         VXM_SCOPE_RC_NO_OBJECT
 */
VxmRc
MozillaScope::getElement(int index, EcmaScriptValue* aValue, ESObject aArray)
{
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSGetElementEvent * evt = new JSGetElementEvent(this, index, aValue, aArray);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleGetElementEvent, MozillaScope::DestroyGetElementEvent);
 
  void * result;
  nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) {
     PL_DestroyEvent(evt);
    return VXM_RC_INTERNAL_ERROR;
  }

  //aValue updated in the event with the value of the array element
  return evt->mRc;
}

void *
MozillaScope::HandleGetElementEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleGetElementEvent );

  JSGetElementEvent * evt = NS_STATIC_CAST(JSGetElementEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  JSObject* obj = NULL;

  // Use the given scope chain
  obj = (JSObject*)evt->mArray;   

  NS_ASSERTION(obj, "Null array in getElement()");

  jsval value;  // Spidermonkey property value

  // Get the Spidermonkey object property
  if (!JS_GetElement(scope->m_Context, obj, evt->mIndex, &value))
  {
    evt->mValue->type = SCOPE_VALUE_UNDEFINED;
    evt->mValue->value.object_val = (void*)NULL;
    evt->mRc = VXM_RC_INTERNAL_ERROR;
    return (void *)evt;
  }

  // Set the output type and value from the returned value
  if (JSVAL_IS_VOID(value))
  {
    evt->mValue->type = SCOPE_VALUE_UNDEFINED;
    evt->mValue->value.object_val = (void*)NULL;
  }

  else if (JSVAL_IS_OBJECT(value))
  {
    evt->mValue->type = SCOPE_VALUE_OBJECT;
    JSObject *temp = 0; 
    JS_ValueToObject(scope->m_Context, value, &temp);
    evt->mValue->value.object_val = temp;
  }
  else if (JSVAL_IS_BOOLEAN(value))
  {
    evt->mValue->type = SCOPE_VALUE_BOOLEAN;
    evt->mValue->value.bool_val = (JSVAL_TO_BOOLEAN(value) != 0);
  }
  else if (JSVAL_IS_NUMBER(value))
  {
    evt->mValue->type = SCOPE_VALUE_NUMBER;
    JS_ValueToNumber(scope->m_Context, value, &evt->mValue->value.num_val);
  }
  else if (JSVAL_IS_STRING(value))
  {
    evt->mValue->type = SCOPE_VALUE_STRING;
    JSString *str = JS_ValueToString(scope->m_Context, value);
    evt->mValue->value.str_val = (wchar_t*)JS_GetStringChars(str);
  }
  else
  {
    evt->mValue->type = SCOPE_VALUE_UNDEFINED;
    evt->mValue->value.object_val = (void*)NULL;
  }

  return (void *)evt;
}

/**Set the value of a field within an object
 * @param index (in) index number of the element in array
 * @param value (in) The new value of the field
 * @param array (in) handle to the element's array
 * @return VXMI_RC_OK, VXMI_RC_INTERNAL_ERROR, VXMI_RC_CANCELED,
 *         VXMI_SCOPE_RC_NO_OBJECT
 */
VxmRc
MozillaScope::setElement(int index, const EcmaScriptValue& aValue, ESObject aArray)
{
  NS_ASSERTION(aArray, "Null array object");

  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSSetElementEvent * evt = new JSSetElementEvent(this, index, aValue, aArray);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleSetElementEvent, MozillaScope::DestroySetElementEvent);
 
  void * result;
  nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) {
     PL_DestroyEvent(evt);
    return VXM_RC_INTERNAL_ERROR;
  }

  return evt->mRc;
}

void *
MozillaScope::HandleSetElementEvent(PLEvent * aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleSetElementEvent );

  JSSetElementEvent * evt = NS_STATIC_CAST(JSSetElementEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  jsval slotVal;

  // Check the input type and set the slot value
  switch (evt->mValue.type)
  {
  case SCOPE_VALUE_BOOLEAN:
    {
      int bVal = (evt->mValue.value.bool_val) ? 1 : 0;
      slotVal = BOOLEAN_TO_JSVAL(bVal);
    }
    break;

  case SCOPE_VALUE_NUMBER:
    {
      jsdouble jdb = ((jsdouble)evt->mValue.value.num_val);
      JS_NewNumberValue(scope->m_Context, jdb, &slotVal);
    }
    break;

  case SCOPE_VALUE_STRING:
    {
      JSString* str = JS_NewUCStringCopyN(scope->m_Context,
                  (jschar*)evt->mValue.value.str_val,
                  wcslen(evt->mValue.value.str_val));
      slotVal = STRING_TO_JSVAL(str);
    }
    break;

  case SCOPE_VALUE_OBJECT:
    {
      JSObject* jobj = ((JSObject*)evt->mValue.value.object_val);
      slotVal = OBJECT_TO_JSVAL(jobj);
    }
    break;

  case SCOPE_VALUE_UNDEFINED:
    {
      slotVal = JSVAL_VOID;
    }
    break;

  default: 
    {
      evt->mRc = VXM_RC_INTERNAL_ERROR;
      return (void *)evt;
    }
  }

  // Set the slot value as a spidermonkey property
  if (!JS_SetElement(scope->m_Context, (JSObject *)evt->mArray, evt->mIndex, &slotVal))
  {
    evt->mRc = VXM_RC_INTERNAL_ERROR;
  }

  return (void *)evt;
}

/**Set the value of a field within an object
 * @param slotName (in) name of the field
 * @param value (in) the new value of the field
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED,
 *         VXM_SCOPE_RC_NO_OBJECT
 */
VxmRc MozillaScope::setSlot(const wchar_t * slotName, const EcmaScriptValue& value)
{
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSSetSlotEvent * evt = new JSSetSlotEvent(this, slotName, value);
  if (!evt) 
  {
     return VXM_RC_INTERNAL_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleSetSlotEvent, MozillaScope::DestroySetSlotEvent);
 
  void * result;
   nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) 
   {
    PL_DestroyEvent(evt);
    return VXM_RC_INTERNAL_ERROR;
  }

  return evt->mRc;
}

void *
MozillaScope::HandleSetSlotEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleSetSlotEvent );

  JSSetSlotEvent * evt = NS_STATIC_CAST(JSSetSlotEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  VxmRc rc = VXM_RC_OK;
  jsval slotVal;

  // Get top of Scope chain
  JSObject* obj = ((SMObject*) scope->m_scopeChain[0])->m_object;

  // return error if scope object not found
  if (obj == NULL)
  {
    evt->mRc =  VXM_RC_INTERNAL_ERROR;
    return (void *)evt;
  }

  // Check the input type and set the slot value
  switch (evt->mValue.type)
  {
  case SCOPE_VALUE_BOOLEAN:
    {
      int bVal = (evt->mValue.value.bool_val) ? 1 : 0;
      slotVal = BOOLEAN_TO_JSVAL(bVal);
    }
    break;

  case SCOPE_VALUE_NUMBER:
    {
      jsdouble jdb = ((jsdouble)evt->mValue.value.num_val);
      JS_NewNumberValue(scope->m_Context, jdb, &slotVal);
    }
    break;

  case SCOPE_VALUE_STRING:
    {
      JSString* str = JS_NewUCStringCopyN(scope->m_Context,
                  (jschar*)evt->mValue.value.str_val,
                  wcslen(evt->mValue.value.str_val));
      slotVal = STRING_TO_JSVAL(str);
    }
    break;

  case SCOPE_VALUE_OBJECT:
    {
      JSObject* jobj = ((JSObject*)evt->mValue.value.object_val);
      slotVal = OBJECT_TO_JSVAL(jobj);
    }
    break;

  case SCOPE_VALUE_UNDEFINED:
    {
      slotVal = JSVAL_VOID;
    }
    break;

  default: 
    {
      evt->mRc = VXM_RC_INTERNAL_ERROR;
      return  (void *)evt;
    }
  }

  // Set the slot value as a spidermonkey property
  if (!JS_SetUCProperty(scope->m_Context,
              obj,
              (jschar*)evt->mSlotName.get(),
              evt->mSlotName.Length(),
              &slotVal))
  {
    evt->mRc = VXM_RC_INTERNAL_ERROR;
  }

  return (void *)evt;
}

/**Set the value of a field within an object
 * @param slotName (in) name of the field
 * @param value (in) the new value of the field
 * @param object (in) handle to the parent object.
 * @return VXMI_RC_OK, VXMI_RC_INTERNAL_ERROR, VXMI_RC_CANCELED
 */
VxmRc
MozillaScope::setSlot(const wchar_t *slotName, const EcmaScriptValue& value, ESObject object)
{
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSSetSlotObjectEvent * evt = new JSSetSlotObjectEvent(this, slotName, value, object);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleSetSlotObjectEvent, MozillaScope::DestroySetSlotObjectEvent);
 
  void * result;
   nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) 
   {
    PL_DestroyEvent(evt);
      return VXM_RC_INTERNAL_ERROR;
  }
  return evt->mRc;
}

void *
MozillaScope::HandleSetSlotObjectEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleSetSlotObjectEvent );

  JSSetSlotObjectEvent * evt = NS_STATIC_CAST(JSSetSlotObjectEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  AutoPushJSContext autopush(scope->m_Context);
  JSAutoRequest autoreq(scope->m_Context);

  VxmRc rc = VXM_RC_OK;
  jsval slotVal;

  JSObject* obj = (JSObject*)evt->mObject;

  // Check the input type and set the slot value
  switch (evt->mValue.type)
  {
  case SCOPE_VALUE_BOOLEAN:
    {
      int bVal = (evt->mValue.value.bool_val) ? 1 : 0;
      slotVal = BOOLEAN_TO_JSVAL(bVal);
    }
    break;

  case SCOPE_VALUE_NUMBER:
    {
      jsdouble jdb = ((jsdouble)evt->mValue.value.num_val);
      JS_NewNumberValue(scope->m_Context, jdb, &slotVal);
    }
    break;

  case SCOPE_VALUE_STRING:
    {
      JSString* str = JS_NewUCStringCopyN(scope->m_Context,
                  (jschar*)evt->mValue.value.str_val,
                  wcslen(evt->mValue.value.str_val));
      slotVal = STRING_TO_JSVAL(str);
    }
    break;

  case SCOPE_VALUE_OBJECT:
    {
      JSObject* jobj = ((JSObject*)evt->mValue.value.object_val);
      slotVal = OBJECT_TO_JSVAL(jobj);
    }
    break;

  case SCOPE_VALUE_UNDEFINED:
    {
      slotVal = JSVAL_VOID;
    }
    break;

  default: 
    return NULL;
  }

  static int count=1;
  printf("->JS_SetUCProperty count = %u\n", count);
  // Set the slot value as a spidermonkey property
  if (!JS_SetUCProperty(scope->m_Context,
              obj,
              (jschar*)evt->mSlotName.get(),
              evt->mSlotName.Length(),
              &slotVal))
  {
    evt->mRc = VXM_RC_INTERNAL_ERROR;
  }
  printf("<-JS_SetUCProperty count = %u\n", count++);

  return (void *)evt;
}

/**Get the recording value of a field within an object
 * @param value (in) the audio recording container
 * @param recHandle (out) returned recording handle of the field, or empty
 *                        if not a recording
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED
 */
VxmRc
MozillaScope::getRecording(const EcmaScriptValue& value, wchar_t** recHandle)
{
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSGetRecordingEvent * evt = new JSGetRecordingEvent(this, value, recHandle);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleGetRecordingEvent, MozillaScope::DestroyGetRecordingEvent);
 
  void * result;
  nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) {
     PL_DestroyEvent(evt);
   }
  return VXM_RC_OK;
}

void *
MozillaScope::HandleGetRecordingEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleGetRecordingEvent );

  VxmRc rc;
  bool bRecording = true;
  EcmaScriptValue result;
    
  JSGetRecordingEvent * evt = NS_STATIC_CAST(JSGetRecordingEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  // check whether value is an object
  if (evt->mValue.type == SCOPE_VALUE_OBJECT)
  {
    // get the value of the VXMLAudioRecording handle property
    rc = scope->getSlot(L"handle", &result, evt->mValue.value.object_val);
    if(rc != VXM_RC_OK)
    {
      evt->mRc = rc;
      return (void *)evt;
    }

    // check whether handle property is a string
    if (result.type == SCOPE_VALUE_STRING)
    {
      // set *recHandle to the handle property
      *evt->mhRecording = new wchar_t[wcslen(result.value.str_val) + 1];
      *evt->mhRecording[0] = 0;
      wcscpy(*evt->mhRecording, result.value.str_val);
    }
    else
    {
      // not a recording
      bRecording = false;
    }
  }
  else
  {
    // not a recording
    bRecording = false;
  }

  if (!bRecording)
  {
    // not a recording, so set *recHandle to be empty
    evt->mhRecording[0] = 0;
  }

  return (void *)evt;
}

/**Set the value of a field within an object to be a handle to a recording
 * @param slotName (in) name of the field
 * @param recHandle (in) returned recording handle of the field
 * @return VXMI_RC_OK, VXMI_RC_INTERNAL_ERROR, VXMI_RC_CANCELED
 */
VxmRc
MozillaScope::setRecording(const wchar_t * slotName, const wchar_t * recHandle)
{
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  if(!eqsvc)
  {
    printf("Failed to get nsIEventQueueService\n");
     return VXM_RC_INTERNAL_ERROR;
  }

   nsCOMPtr<nsIEventQueue> eventQ;
   eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                     getter_AddRefs(eventQ));
   if (!eventQ) 
  {
     return VXM_RC_INTERNAL_ERROR;
   }

  JSSetRecordingEvent * evt = new JSSetRecordingEvent(this, slotName, recHandle);
  if (!evt) 
  {
     return VXM_RC_MEM_ALLOC_ERROR;
   }
 
   PL_InitEvent(evt, this, MozillaScope::HandleSetRecordingEvent, MozillaScope::DestroySetRecordingEvent);
 
  void * result;
  nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
   if (NS_FAILED(rv)) {
     PL_DestroyEvent(evt);
   }
  return VXM_RC_OK;
}

void *
MozillaScope::HandleSetRecordingEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE( MozillaScope::HandleSetRecordingEvent );

  JSSetRecordingEvent * evt = NS_STATIC_CAST(JSSetRecordingEvent *, aEvent);
  MozillaScope * scope = NS_STATIC_CAST(MozillaScope *, aEvent->owner);

  EcmaScriptValue result;
  VxmRc rc;

  // Create a new object of type VXMLAudioRecording
  NS_NAMED_LITERAL_STRING( begin, "new VXMLAudioRecording('" );
  NS_NAMED_LITERAL_STRING( end, "');" );
  nsAutoString audioRecordingObj;
  audioRecordingObj += begin;
  audioRecordingObj += evt->mhRecording;
  audioRecordingObj += end;
  rc = scope->eval((wchar_t *)audioRecordingObj.get(), &result, false);  // returns an object
  if (rc != VXM_RC_OK)
  {
    evt->mRc = rc;
    return (void *)evt;
  }

  // make sure result was an object
  if (result.type != SCOPE_VALUE_OBJECT)
  {
    evt->mRc = VXM_RC_INTERNAL_ERROR;
    return (void *)evt;
  }

  // set slotname property of object to the VXMLAudioRecording object
  rc = scope->setSlot((wchar_t *)evt->mSlotName, result);
  if(rc != VXM_RC_OK)
  {
    evt->mRc = VXM_RC_INTERNAL_ERROR;
  }

  return (void *)evt;
}

/**Delete a string that was created from a previous eval(), getSlot() or getElement() call
 * @param ptrString (in) pointer to the string to free
 * @return VXM_RC_OK, VXM_RC_INTERNAL_ERROR, VXM_RC_CANCELED
 */
VxmRc MozillaScope::freeString(wchar_t * ptrString)
{
  return VXM_RC_OK;
}

void 
MozillaScope::DestroyInitEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroyEvalEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroyEnterEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroySetSlotEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroyGetSlotEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroySetSlotObjectEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroyGetSlotObjectEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroyCreateObjectEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroyCreateArrayObjectEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroyLeaveEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroyGetElementEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroySetElementEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroyGetRecordingEvent(PLEvent *evt) {/* delete evt; */}

void 
MozillaScope::DestroySetRecordingEvent(PLEvent *evt) {/* delete evt; */}

